
#include "math.h"
#include "truetypeout.h"
#include <fstream>

extern "C"
{
	void TextOutline_ClearAll();
	void TextOutline_Clear();
	void TextOutline_AddLine(int x1, int y1, int x2, int y2);
	void TextOutline_AddBezier(int x1, int y1, int x2, int y2, int x3, int y3);
	void TextOutline_AddWire();
	void TextOutline_MakeFace();

	int TextOutline_PreviewRecord(char* str, int type, double x, double y, double z, double s);
	int TextOutline_AddToDB(int type);
	void TextOutline_Record(char* str, int bClearTemp);
	void TextOutline_ClearRecordTempFaces();
	void TextOutline_ClearAllRecord();
	int TextOutline_AddRecord(char* str, int type, double x, double y, double z, double s);
}

extern wxWindow* GetMainGLWindow();
static const int SAMPLE_FONT_HT = 48; // font height for ChooseFont dialog
static const int GGO_FONT_HT = 576;   // font height for glyph realizing

#ifdef WIN32

static POINTFX AvgPointsFX(  const POINTFX p,  const POINTFX q);

static bool DrawQSpline(  HDC hDC,  const POINTFX p1,  const POINTFX p2,  const POINTFX p3);

static inline FIXED FixedFromDouble(const double d);

static POINTFX AvgPointsFX(  const POINTFX p,  const POINTFX q)
{
	long x, y;            // average of x's and y's, as longs
	POINTFX r;            // average, converted back to POINTFX
	// Convert coords to long and average.
	x = (*((long far *)&(p.x)) + *((long far *)&(q.x))) >> 1;
	y = (*((long far *)&(p.y)) + *((long far *)&(q.y))) >> 1;
	r.x = *(FIXED *)&x;
	r.y = *(FIXED *)&y;
	return (r);
}

static inline FIXED FixedFromDouble(const double d)
{
	long l;  
	l = (long) (d * 65536L);
	return *(FIXED *)&l;
}

TrueTypeOutline::TrueTypeOutline()
{
	// A constructor should initialize all data members.
	m_bRealized = false;
	m_cbBuffer = 0;
	m_dAngleDegrees = 0;
	memset(&m_gm, 0, sizeof(m_gm));
	m_lpvBuffer = NULL;
	memset(&m_mat2, 0, sizeof(m_mat2));
	m_uChar = 0;
	// Set scale = 75% to fit into client area initially.
	m_dScale = 0.75;

	m_xExtent = 10.0;
	m_yExtent = 10.0;
	m_zExtent = 10.0;

  memset(&m_lf, 0, sizeof(m_lf));
  m_lf.lfPitchAndFamily = FF_ROMAN;
  m_lf.lfHeight         = GGO_FONT_HT;
  m_lf.lfCharSet        = ANSI_CHARSET;
  m_lf.lfOutPrecision   = OUT_TT_PRECIS;
  strcpy(m_lf.lfFaceName, "Times New Roman");
}

TrueTypeOutline::~TrueTypeOutline()
{
	Unrealize();
}                  

void TrueTypeOutline::GetOutline(const int xOffset, const int yOffset, std::vector<int>& types, std::vector<POINT>& pnts) const
{
	types.clear();
	pnts.clear();

	LPTTPOLYGONHEADER lpttph;  // pointer to current polygon in glyph
	LPTTPOLYCURVE lpttpc;      // pointer to current polycurve in polygon
	DWORD dwHeaderOffset;      // byte offset of current polygon header
	//     from start of buffer
	DWORD dwCurveOffset;       // byte offset of current polycurve header
	//     from start of polygon header
	DWORD dwStructSize;        // size of current polycurve (depends on
	//     the number of points in the polycurve)
	POINT ptPolyStart;         // polygon start point
	POINT ptCurveLast;         // last point on previous curve
	POINT pt;                  // temporary holder of point coordinates
	int i;                     // to loop thru lines or splines
	POINTFX p1, p2, p3;        // three points defining spline

	if (!IsRealized())
		return;

	assert(m_lpvBuffer != NULL);

	// Outer while loops for all polygon headers (can be more than one).
	dwHeaderOffset = 0;
	while (m_cbBuffer >= (dwHeaderOffset + sizeof(TTPOLYGONHEADER)))
	{
		// Get pointer to start of the polygon.
		lpttph = (LPTTPOLYGONHEADER)(((char *)m_lpvBuffer) + dwHeaderOffset);
		assert(lpttph->dwType == TT_POLYGON_TYPE);
		// Convert polygon's start point to pixels.
		FromGGOPoint(&ptPolyStart, lpttph->pfxStart, xOffset, yOffset);
		// Init last point of last record to polygon start point.
		ptCurveLast = ptPolyStart;
		// Position graphics current pointer to start of polygon.
		//    MoveToEx(hDC, ptPolyStart.x, ptPolyStart.y, NULL);

		dwCurveOffset = sizeof(TTPOLYGONHEADER);
		while (lpttph->cb >= (dwCurveOffset + sizeof(TTPOLYCURVE)))
		{
			// Get pointer to start of polycurve.
			lpttpc = (LPTTPOLYCURVE)(((char *)m_lpvBuffer) + dwHeaderOffset
				+ dwCurveOffset);

			// Test record type, draw polyline or series of Beziers accordingly.
			switch (lpttpc->wType)
			{
			case TT_PRIM_LINE:
				// Draw polyline connecting GGO points (there are lpttpc->cpfx).
				for (i = 0; i < lpttpc->cpfx; i++)
				{
					FromGGOPoint(&pt, lpttpc->apfx[i], xOffset, yOffset);
					//            LineTo(hDC, pt.x, pt.y);
					types.push_back(1); //add one line
					pnts.push_back(ptCurveLast);
					pnts.push_back(pt);
					ptCurveLast = pt;
				}
				ptCurveLast = pt;
				break;
			case TT_PRIM_QSPLINE:
				// Draw series of Beziers connecting GGO points.
				// But for initial Bezier, grab last point on previous curve.
				p3.x.value = (short)ptCurveLast.x;
				p3.x.fract = 0;
				p3.y.value = (short)ptCurveLast.y;
				p3.y.fract = 0;
				// Draw QSplines based on GGO points (there are lpttpc->cpfx-1).
				for (i = 0; i < lpttpc->cpfx - 1; i++)
				{
					// p1 is 1st control -- last point in this or previous contour.
					p1 = p3;
					// p2 is handle -- a point in the record.
					FromGGOPoint(&pt, lpttpc->apfx[i], xOffset, yOffset);
					p2.x.value = (short)pt.x;
					p2.x.fract = 0;
					p2.y.value = (short)pt.y;
					p2.y.fract = 0;
					// p3 is 2nd control -- the midpoint of 2 spline points
					// except for the last spline, when it is the second to
					// the last point in the spline record.
					p3 = (i == (lpttpc->cpfx-2)) ? lpttpc->apfx[i+1] :
						AvgPointsFX(lpttpc->apfx[i], lpttpc->apfx[i+1]);
					FromGGOPoint(&pt, p3, xOffset, yOffset);
					p3.x.value = (short)pt.x;
					p3.x.fract = 0;
					p3.y.value = (short)pt.y;
					p3.y.fract = 0;
					// Call deCasteljau to draw quadratic Bezier curve.
					types.push_back(2); //bezier
					POINT p;
					p.x = p1.x.value;
					p.y = p1.y.value;
					pnts.push_back(p);
					p.x = p2.x.value;
					p.y = p2.y.value;
					pnts.push_back(p);
					p.x = p3.x.value;
					p.y = p3.y.value;
					pnts.push_back(p);
				}
				ptCurveLast = pt;
				break;
			default:
				assert(false); // must be one of the two cases above
				break;
			} // switch

			// Increment curve offset so point to next polycurve.
			dwStructSize = sizeof(TTPOLYCURVE) + ((lpttpc->cpfx - 1)
				* sizeof(POINTFX));
			dwCurveOffset += dwStructSize;
		} // inner while (polycurve)
		// Close polygon by drawing a line back to the start point.
		//    LineTo(hDC, ptPolyStart.x, ptPolyStart.y);
		if(abs(ptCurveLast.x-ptPolyStart.x)>0 || abs(ptCurveLast.y-ptPolyStart.y)>0)
		{
			types.push_back(1);
			pnts.push_back(ptCurveLast);
			pnts.push_back(ptPolyStart);
		}
		types.push_back(-1);

		// Increment polygon offset so point to next polygon.
		dwHeaderOffset += lpttph->cb;
	} 
}

inline void TrueTypeOutline::FromGGOPoint(POINT * const ppt,	const int xGGO, 	const int yGGO, const int xOffset, 	const int yOffset) const
{
	ppt->x = static_cast<long>
		(m_dScale * (xGGO - m_gm.gmptGlyphOrigin.x) + xOffset);
	ppt->y = static_cast<long>
		(m_dScale * (yGGO-m_gm.gmptGlyphOrigin.y ) + yOffset);
}

inline void TrueTypeOutline::FromGGOPoint(POINT * const ppt,	const POINTFX & ptfxGGO,  const int xOffset,  const int yOffset ) const
{
	FromGGOPoint(ppt, ptfxGGO.x.value, ptfxGGO.y.value, xOffset, yOffset);
}

void TrueTypeOutline::FromGGORect( RECT * const prect, const int nGGOLeft, const int nGGOTop, const int nGGORight,  const int nGGOBottom, const int xOffset,  const int yOffset) const
{
	POINT pt;

	FromGGOPoint(&pt, nGGOLeft, nGGOTop, xOffset, yOffset);
	prect->left = pt.x;
	prect->top = pt.y;

	FromGGOPoint(&pt, nGGORight, nGGOBottom, xOffset, yOffset);
	prect->right = pt.x;
	prect->bottom = pt.y;
}

double TrueTypeOutline::GetAngle()
{
	return m_dAngleDegrees;
}

void TrueTypeOutline::GetBoundingRect( RECT * const prect, const int xOffset, const int yOffset) const
{
	if (IsRealized())
	{
		FromGGORect(prect,  m_gm.gmptGlyphOrigin.x,   m_gm.gmptGlyphOrigin.y,
			m_gm.gmptGlyphOrigin.x + m_gm.gmBlackBoxX,  m_gm.gmptGlyphOrigin.y - m_gm.gmBlackBoxY,    xOffset,    yOffset);
	}
	else
	{
		prect->left = 0;
		prect->top = 0;
		prect->right = 0;
		prect->bottom = 0;
	}
}

bool TrueTypeOutline::IsRealized() const
{
	return m_bRealized;
}

void TrueTypeOutline::MakeRotationMatrix()
{
	double cosTheta, sinTheta;

	cosTheta = cos(PI180*m_dAngleDegrees);
	sinTheta = sin(PI180*m_dAngleDegrees);

	m_mat2.eM11 = FixedFromDouble(cosTheta);
	m_mat2.eM12 = FixedFromDouble(sinTheta);
	m_mat2.eM21 = FixedFromDouble(-sinTheta);
	m_mat2.eM22 = FixedFromDouble(cosTheta);
} 

int TrueTypeOutline::SetGlyphFontByDialog() 
{
  CHOOSEFONT cf;             // needed by ChooseFont common dialog
  BOOL bResult;              // return value of ChooseFont
  LOGFONT curFont;

  memcpy(&curFont, &m_lf, sizeof(LOGFONT));
  // Start all CHOOSEFONT fields to 0.
  memset(&cf, 0, sizeof(cf));
  cf.lStructSize = sizeof(cf);
//  cf.hwndOwner   = pView->m_hWnd;
  // Init logFont from member variable holding current settings.
  cf.lpLogFont   = &(curFont);
  // Flags:  Use logfont structure to initialize the dialog controls,
  // show only screen fonts, show only True Type fonts.
  cf.Flags = CF_INITTOLOGFONTSTRUCT | CF_SCREENFONTS | CF_TTONLY;
  
  // Set font height so sample text appears in dialog.
  curFont.lfHeight = SAMPLE_FONT_HT;
  
  // Display the ChooseFont common dialog.
  bResult = ChooseFont(&cf);
  
  // Reset font height to GGO_FONT_HT before realizing the character -- this
  // is the size for GGO to use when creating the glyph outline.
  m_lf.lfHeight = GGO_FONT_HT;
  // If ChooseFont failed, then exit.
  if (!bResult)
    return 0;
  memcpy(&m_lf, &curFont, sizeof(LOGFONT));
  m_lf.lfHeight = GGO_FONT_HT;
  return 1;
}

bool TrueTypeOutline::Realize()
{
	DWORD dwResult;
	wxWindow* pWindow = GetMainGLWindow();
	if(!pWindow)
		return false;
	wxWindowDC dc(pWindow);
	HDC hDC = (HDC)dc.GetHDC();

	/*
	LOGFONT m_lf;
	memset(&m_lf, 0, sizeof(m_lf));
	m_lf.lfPitchAndFamily = FF_ROMAN;
	m_lf.lfHeight         = GGO_FONT_HT;
	m_lf.lfCharSet        = ANSI_CHARSET;
	m_lf.lfOutPrecision   = OUT_TT_PRECIS;
	strcpy(m_lf.lfFaceName, "Times New Roman");
	strcpy(m_lf.lfFaceName, font.GetFaceName().char_str());
	*/

	HFONT hFont = CreateFontIndirect(&m_lf);
	HFONT hOldFont = (HFONT) SelectObject(hDC, hFont);

	if (m_bRealized)
		Unrealize();

	MakeRotationMatrix();

	m_format = GGO_NATIVE;
	m_cbBuffer = GetGlyphOutlineW(hDC, m_uChar, m_format, &m_gm, 0, NULL, &m_mat2);
	if (m_cbBuffer == GDI_ERROR)
	{
		Unrealize();
	    ReselectOldFont(hDC, hOldFont, hFont);		
		return false;
	} 

	m_lpvBuffer = malloc(m_cbBuffer);
	if (m_lpvBuffer == NULL)
	{
		Unrealize();
		ReselectOldFont(hDC, hOldFont, hFont);
		return false;
	}

	dwResult = GetGlyphOutlineW(hDC, m_uChar, m_format, &m_gm, m_cbBuffer,   m_lpvBuffer, &m_mat2);
	if (dwResult == GDI_ERROR)
	{
		Unrealize();
		ReselectOldFont(hDC, hOldFont, hFont);
		return false;
	}
    ReselectOldFont(hDC, hOldFont, hFont);
	m_bRealized = true;
	return true;
}

bool TrueTypeOutline::Realize(  const UINT uChar)
{
	m_uChar = uChar;
	return Realize();
}

void TrueTypeOutline::SetAngle(const double dNewValue)
{
	long lIntegerPart = static_cast<long>(dNewValue);
	double dFractionPart = dNewValue - lIntegerPart;

	m_dAngleDegrees = (lIntegerPart % 360) + dFractionPart;

	if (m_dAngleDegrees < 0)
		m_dAngleDegrees += 360;
}

void TrueTypeOutline::ReselectOldFont(const HDC hDC, const HFONT hOldFont, const HFONT hFont)
{
  SelectObject(hDC, hOldFont);
  DeleteObject(hFont);
}

void TrueTypeOutline::Unrealize()
{
	memset(&m_gm, 0, sizeof(m_gm));
	m_cbBuffer = 0;
	free(m_lpvBuffer); // safe even if m_lpvBuffer is NULL
	m_lpvBuffer = NULL;
	m_bRealized = false;
}
#endif
#ifdef WIN32
TrueTypeOutline *s_TextOutlineSolver=NULL;
#endif
void TrueTypeOutline_Create()
{
#ifdef WIN32
	if(s_TextOutlineSolver)
		return;
	s_TextOutlineSolver = new TrueTypeOutline();
#endif
}
int TrueTypeOutline_SetFont()
{
#ifdef WIN32
	TrueTypeOutline_Create();
	if(s_TextOutlineSolver->SetGlyphFontByDialog())
	{
		TextOutline_ClearAllRecord();
		return 1;
	}
	else
		return 0;
#else
	return 0;
#endif
}
//type 0 curve  1 surface  2 solid
extern "C"
{
	void TrueTypeOutline_GetOutline(char* rawS, int type, int bPreview, double xExt, double yExt, double zExt, double meshSize)
	{
	#ifdef WIN32
		wxString s = rawS;
		int  j;
		TextOutline_ClearAll();
		int xOffset=0, yOffset= 0, offset=0, yMax=0;
		std::vector<std::vector<int> > typess;
		std::vector<std::vector<POINT> > pntss;
		std::vector<int> yOffsets;
		RECT rect;
		UINT c;

		TrueTypeOutline_Create();
		TextOutline_ClearAll();
		TextOutline_ClearRecordTempFaces();
		if(bPreview && TextOutline_PreviewRecord(s.char_str(), type, xExt, yExt, zExt, meshSize))
		{
			return;
		}
		if(!bPreview && TextOutline_AddRecord(s.char_str(), type, xExt, yExt, zExt, meshSize))
		{
			return;
		}
		xOffset = 0;
		yOffset = 0;

		std::wstring str ( (wchar_t * )s.wchar_str());
		std::wstring::iterator i;
		for(i=str.begin(); i!=str.end(); ++i)
		{
			c = (UINT)(*i);
			s_TextOutlineSolver->Realize(c);
			s_TextOutlineSolver->GetBoundingRect(&rect, 0, 0);

			std::vector<int> types;
			std::vector<POINT> pnts;
			yOffset = -rect.bottom+rect.top;
			s_TextOutlineSolver->GetOutline(xOffset, yOffset,  types, pnts);
			if(s_TextOutlineSolver->m_gm.gmCellIncY>yMax)
				yMax = s_TextOutlineSolver->m_gm.gmCellIncY;
			typess.push_back(types);
			pntss.push_back(pnts);
			yOffsets.push_back(s_TextOutlineSolver->m_gm.gmBlackBoxY);
			xOffset+=s_TextOutlineSolver->m_gm.gmCellIncX;
		}
		for(int k=0; k<(int)typess.size(); ++k)
		{
			std::vector<int>& types = typess[k];
			std::vector<POINT>& pnts=pntss[k];
//			yOffset = yMax-yOffsets[k];
			offset = 0;
			yOffset = 0;
			for(j=0; j<types.size(); ++j)
			{
				if(types[j]==-1)
				{
					TextOutline_AddWire();
					continue;
				}
				else if(types[j]==1)
				{
					TextOutline_AddLine(pnts[offset].x, pnts[offset].y+yOffset, pnts[offset+1].x, pnts[offset+1].y+yOffset);
					offset+=2;
				}
				else if(types[j]==2)
				{
					TextOutline_AddBezier(pnts[offset].x, pnts[offset].y+yOffset, pnts[offset+1].x, pnts[offset+1].y+yOffset, pnts[offset+2].x, pnts[offset+2].y+yOffset);
					offset+=3;
				}
			}
			TextOutline_MakeFace();
//			xOffset+=s_TextOutlineSolver->m_gm.gmCellIncX;
		}
		TextOutline_Record(s.char_str(), 1);
		TextOutline_ClearAll();

		if(!bPreview)
		{
			TextOutline_AddRecord(s.char_str(), type, xExt, yExt, zExt, meshSize);
		}
//			TextOutline_AddToDB(type);
		else
		{
			TextOutline_PreviewRecord(s.char_str(), type, xExt, yExt, zExt, meshSize);
			/*
			if(type==0)
				TextOutline_PreviewWires(xExt);
			else if(type==1)
				TextOutline_PreviewFace(xExt);
			else if(type==2)
				TextOutline_PreviewSolid(xExt, zExt);
				*/
		}
	#endif
	}
}
//#endif
